/*
 * Author:	William Chia-Wei Cheng (william@cs.ucla.edu)
 *
 * Copyright (C) 1990, 1991, William Cheng.
 */
#ifndef lint
static char RCSid[] =
      "@(#)$Header: /tmp_mnt/n/kona/tangram/u/william/X11/TGIF2/RCS/cursor.c,v 2.0 91/03/05 12:46:56 william Exp $";
#endif

#include <X11/Xlib.h>
#include <X11/cursorfont.h>
#include "const.h"
#include "types.h"

#include "choice.e"
#include "setup.e"

#include "xbm/null.xbm"
#include "xbm/nullmask.xbm"
#include "xbm/text_cur_image.xbm"

Cursor	nullCursor;
Cursor	cornerCursor;
Cursor	defaultCursor;
Cursor	handCursor;

static GC	textCursorGC;
static Cursor	textCursor;
static Pixmap	textPixmap;
static Pixmap	nullPixmap;
static Pixmap	nullMaskPixmap;

static XImage	* textCursorImage;

void SetTextCursor (window)
   Window	window;
{
   XDefineCursor (mainDisplay, window, textCursor);
}

void SetNullCursor (window)
   Window	window;
{
   XDefineCursor (mainDisplay, window, nullCursor);
}

void SetDefaultCursor (window)
   Window	window;
{
   XUndefineCursor (mainDisplay, window);
}

void ShowCursor ()
{
   if (curChoice == DRAWTEXT)
      SetTextCursor (drawWindow);
   else
      SetDefaultCursor (drawWindow);
}

void CreateCursor ()
{
   XGCValues	values;
   XColor	color;

   textPixmap = XCreateBitmapFromData (mainDisplay, mainWindow,
         text_cur_image_bits, text_cur_image_width, text_cur_image_height);

   values.foreground = myFgPixel;
   values.background = myBgPixel;
   values.fill_style = FillOpaqueStippled;
   values.stipple = textPixmap;
   textCursorGC = XCreateGC (mainDisplay, mainWindow,
         GCForeground | GCBackground | GCFillStyle | GCStipple, &values);

   textCursorImage = XCreateImage (mainDisplay, mainVisual, 1, XYBitmap, 0,
         text_cur_image_bits, text_cur_image_width, text_cur_image_height, 8,
         2);
   textCursorImage->byte_order = LSBFirst;
   textCursorImage->bitmap_bit_order = LSBFirst;
   textCursorImage->data = text_cur_image_bits;

   textCursor = XCreateFontCursor (mainDisplay, XC_xterm);
   cornerCursor = XCreateFontCursor (mainDisplay, XC_ul_angle);
   defaultCursor = XCreateFontCursor (mainDisplay, XC_arrow);
   handCursor = XCreateFontCursor (mainDisplay, XC_hand2);

   nullPixmap = XCreatePixmap (mainDisplay, mainWindow, null_width,
         null_height, 1);
   nullMaskPixmap = XCreatePixmap (mainDisplay, mainWindow, nullmask_width,
         nullmask_height, 1);
   nullCursor = XCreatePixmapCursor (mainDisplay, nullPixmap, nullMaskPixmap,
         &color, &color, 0, 0);
}

void PutCursor (window, x, y, foreground)
   Window	window;
   int		x, y, foreground;
{
   XSetForeground (mainDisplay, textCursorGC, foreground);
   XPutImage (mainDisplay, window, textCursorGC, textCursorImage, 0, 0, x, y,
         text_cur_image_width, text_cur_image_height);
}

void CleanUpCursors ()
{
   XFreePixmap (mainDisplay, textPixmap);
   XFreeGC (mainDisplay, textCursorGC);
/* XDestroyImage (textCursorImage); */

   XFreeCursor (mainDisplay, textCursor);
   XFreeCursor (mainDisplay, cornerCursor);
   XFreeCursor (mainDisplay, defaultCursor);
   XFreeCursor (mainDisplay, handCursor);

   XFreePixmap (mainDisplay, nullPixmap);
   XFreePixmap (mainDisplay, nullMaskPixmap);
   XFreeCursor (mainDisplay, nullCursor);
}
